unit XRichEdt;

interface

uses
  KeyWords,
  SysUtils, WinProcs, Classes, Comctrls, RichEdit, Graphics, Messages;

type
  TXRichEdt= class(TRichEdit)
  private
    { Interne Variablen }

    FFilename        : string;
    FLetters         : set of AnsiChar;{ Menge der Buchstaben, aus denen ein Schlsselwort
                                         bestehen kann. }
    FSyntax          : Boolean;        { Gibt an, ob Schlsselwrter der Listen 'FKeyWords' und
                                         'FKeyWords2' hervorgehoben werden. }
    FKeyFont1,
    FKeyFont2        : TFont;          { Font fr die Schlsselwrter der Listen 'FKeyWords1'
                                         bzw. 'FKeyWords2', sofern sie hervorgehoben werden
                                         sollen. }
    FUseFullFont     : Boolean;        { Gibt an, ob fr die Syntaxhervorhebung die gesamten
                                         KeyFonts benutzt werden oder nur Farbe+Stil (ist etwas
                                         schneller. }
    FKeyWords1,
    FKeyWords2       : TKeyword;       { Listen der hervorzuhebenden Schlsselwrter }

    Coloring         : Boolean;        { Gibt an, ob der Einfrbevorgang luft. }
    Pos_alt, Pos_neu : Integer;        { Alter und aktueller Wert von 'SelStart'; wird verwendet
                                         um bei nderungen am Inhalt festzustellen, welcher
                                         Bereich neu eingefrbt werden muss. }
    TopLine_alt      : Integer;        { Bisherige Nummer der ersten sichtbaren Zeile; wird
                                         fr das Ereignis 'OnTopLineChange' verwendet. }
    FOnTopLineChange : TNotifyEvent;

    { Zugriffsmethoden fr Eigenschaften }

    procedure SetFilename(Name:string);
    function GetLetters: string;
    procedure SetLetters(Value:string);
    procedure SetKeyFont1(Value:TFont);
    procedure SetKeyFont2(Value:TFont);
    procedure SetUseFullFont(Value:Boolean);
    procedure SetKeyWords1 (KW:TKeyWord);
    procedure SetKeyWords2 (KW:TKeyWord);

    { Methoden fr die Ereignisbehandlung }

    procedure WMMOUSEWHEEL(var Message:TMessage); message WM_MOUSEWHEEL;
    procedure WMVSCROLL(var Message:TMessage); message WM_VSCROLL;
    procedure TopLineChange;

    { Interne Funktionen }

    procedure FormatRange(Pos,Len:Integer; Font:TFont);
                                       { Angegebenen Bereich entsprechend 'Font' formatieren. }
    procedure Bereich_faerben(Pos,Len:Integer);
                                       { Angegebenen Bereich entsprechend der Schlsselwrter
                                         und Zeichenstze formatieren. }
  protected
    procedure SelectionChange; override;
    procedure Change; override;
    procedure Loaded; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    property Filename: string
             read FFilename write SetFilename;
    procedure LoadFromFile(const Name:string); virtual;
    procedure SaveToFile(const Name:string); virtual;
    function GetCurrentWord : string;
    function GetCaretPosition(var Col:Integer) : Integer;
  published
    property Letters: string
             read GetLetters write SetLetters;
    property KeyWords1: TKeyWord
             read FKeyWords1 write SetKeyWords1;
    property KeyWords2: TKeyWord
             read FKeyWords2 write SetKeyWords2;
    property UseFullFont: Boolean
             read FUseFullFont write SetUseFullFont default False;
    property Syntaxhervorhebung: Boolean
             read FSyntax write FSyntax default False;
    property KeyFont1: TFont
             read FKeyFont1 write SetKeyFont1;
    property KeyFont2: TFont
             read FKeyFont2 write SetKeyFont2;
    property OnTopLineChange: TNotifyEvent
             read FOnTopLineChange write FOnTopLineChange;
  end;

procedure Register;
  
implementation

procedure Register;
begin
  RegisterComponents('Beispiele',[TXRichEdt]);
end;

{ -------------------- Erzeugen und Lschen der Komponente }

constructor TXRichEdt.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FFilename    := '';
  FLetters     := ['a'..'z','A'..'Z'];
  FKeyWords1   := TKeyWord.Create;
  FKeyWords2   := TKeyWord.Create;
  FUseFullFont := False;
  FSyntax      := False;
  FKeyFont1    := TFont.Create;
  FKeyFont2    := TFont.Create;
  Pos_alt      := -1;
  Pos_neu      := 0;
  Coloring     := False;
  TopLine_alt  := 0;
end;


destructor TXRichEdt.Destroy;
begin
  KeyFont1.Free;
  KeyFont2.Free;
  KeyWords1.Free;
  KeyWords2.Free;
  inherited Destroy;
end;


{ -------------------- Zugriffsmethoden fr Eigenschaften }

procedure TXRichEdt.SetFilename(Name:string);
var
  ext_alt, ext_neu: string;
begin
  if CompareText(Name, FFileName) <> 0 then begin
    ext_alt := ExtractFileExt(FFilename);
    ext_neu := ExtractFileExt(Name);
    FFilename := Name;
    if CompareText(ext_alt,ext_neu)<>0 then begin
      KeyWords1.SetRange(ext_neu);
      KeyWords2.SetRange(ext_neu);
      Bereich_faerben(0,Length(Text));
      ClearUndo;
    end;
  end;
end;


function TXRichEdt.GetLetters: string;
var
  Ch: AnsiChar;
begin
  result := '';
  for Ch := Low(AnsiChar) to High(AnsiChar) do
    if (Ch in FLetters) and not (Ch in ['a'..'z','A'..'Z']) then
      result := result + Ch;
end;


procedure TXRichEdt.SetLetters(Value:string);
var
  i: Integer;
begin
  FLetters := ['a'..'z','A'..'Z'];
  for i := 1 to Length(Value) do
    FLetters := FLetters + [Value[i]];
end;


procedure TXRichEdt.SetKeyFont1(Value:TFont);
begin
  FKeyFont1.Assign(Value);
  Bereich_faerben(0,Length(Text));
  ClearUndo;
end;


procedure TXRichEdt.SetKeyFont2(Value:TFont);
begin
  FKeyFont2.Assign(Value);
  Bereich_faerben(0,Length(Text));
  ClearUndo;
end;


procedure TXRichEdt.SetUseFullFont(Value:Boolean);
begin
  if FUseFullFont <> Value then begin
    FUseFullFont := Value;
    Bereich_faerben(0,Length(Text));
    ClearUndo;
  end;
end;


procedure TXRichEdt.SetKeywords1(KW:TKeyWord);
begin
  KeyWords1.Assign(KW);
  Bereich_faerben(0,Length(Text));
  ClearUndo;
end;


procedure TXRichEdt.SetKeywords2(KW:TKeyWord);
begin
  KeyWords2.Assign(KW);
  Bereich_faerben(0,Length(Text));
  ClearUndo;
end;


{ -------------------- Methoden fr die Ereignisbehandlung }

procedure TXRichEdt.WMMOUSEWHEEL(var Message:TMessage);
var
  TopLine_neu: Integer;
begin
  inherited;
  TopLine_neu := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
  if TopLine_neu<>TopLine_alt then begin
    TopLine_alt := TopLine_neu;
    TopLineChange;
  end;
end;


procedure TXRichEdt.WMVSCROLL(var Message:TMessage);
begin
  inherited;
  TopLineChange;
end;


procedure TXRichEdt.TopLineChange;
begin
  if Assigned(OnTopLineChange) then OnTopLineChange(Self);
end;



{ -------------------- Interne Methoden }

{ ---------- Angegebenen Bereich entsprechend 'Font' formatieren. }

procedure TXRichEdt.FormatRange(Pos,Len:Integer; Font:TFont);
var
  CharRange: TCharRange;
begin
  CharRange.cpMin := Pos;
  CharRange.cpMax := Pos + Len;
  SendMessage(Handle, EM_EXSETSEL, 0, Longint(@CharRange));

  if FUseFullFont then
    SelAttributes.Assign(Font)
  else begin
    { Nur Farbe und Stil bernehmen ist deutlich schneller }
    SelAttributes.Color := Font.Color;
    SelAttributes.Style := Font.Style;
  end;
end;


{ ----------Angegebenen Bereich entsprechend der Schlsselwrter und Zeichenstze formatieren.
            Wrter, die unmittelbar vor oder hinter dem angegebenen Bereich liegen, werden
            noch "mitgenommen". }

procedure TXRichEdt.Bereich_faerben(Pos,Len:Integer);
var
  MaxLen, p, q : Integer;
  KW: string;
  Inhalt: AnsiString;
begin
  if not Syntaxhervorhebung then Exit;

  Inhalt := Text;
  MaxLen := Length(Inhalt);
  { Falls 'Len' zu gro ist, dann wird dieser Wert reduziert. }
  if (Pos>MaxLen) or (Pos<0) then
    Exit
  else if Pos+Len>MaxLen then
    Len := MaxLen - Pos;

  { Das erste Zeichen des zu markierenden Bereichs ist 'Inhalt[Pos+1]'. Wenn unmittelbar davor
    Buchstaben stehen, dann sollen diese noch einbezogen werden. }
  p := Pos;
  while (Pos>0) and (Inhalt[Pos] in FLetters) do Dec(Pos);
  Inc(Len, p-Pos);
  { Das letzte Zeichen des zu markierenden Bereichs ist 'Inhalt[Pos+Len]'. Wenn unmittelbar
    danach Buchstaben stehen, dann sollen diese auch noch einbezogen werden. }
  p := Pos+Len;
  while (p<MaxLen) and (Inhalt[p+1] in FLetters) do Inc(p);
  Inc(Len, p-Pos-Len);

  if Len<=0 then Exit;

  { Schriftart fr den zu markierenden Bereich zunchst auf Standard zurcksetzen }
  Coloring := True;
  SendMessage(Handle, EM_HIDESELECTION, LongInt(True), LongInt(False));
  FormatRange(Pos,Len,Font);

  { Jetzt den Bereich durchlaufen und nach Schlsselwrtern suchen }
  p := Pos;
  while p<Pos+Len do begin
    { Weitergehen bis zum nchsten Buchstaben }
    while (p<Pos+Len) and not (Inhalt[p+1] in FLetters) do Inc(p);
    if p<Pos+Len then begin
      q := p+1;
      while (q<Pos+Len) and (Inhalt[q+1] in FLetters) do Inc(q);
      KW := Copy(Inhalt,p+1,q-p);
      if KeyWords1.IsKeyword(KW) then
        FormatRange(p, q-p, KeyFont1)
      else if KeyWords2.IsKeyword(KW) then
        FormatRange(p, q-p, KeyFont2);
      p := q+1;
    end;
  end;

  Coloring := False;
  SendMessage(Handle, EM_HIDESELECTION, LongInt(False), LongInt(False));
end;


{ -------------------- Methoden fr die Ereignisbehandlung }

{ ---------- Bei nderung der Auswahl wird die alte Cursorposition gespeichert, dadurch
             kann spter bei einer nderung der neu einzufrbende Textteil eingeschrnkt
             werden. }

procedure TXRichEdt.SelectionChange;
var
  TopLine_neu: Integer;
begin
  if not Coloring then begin
    { Prfen, ob sich die erste sichtbare Zeile gendert hat; wenn ja, dann das Ereignis
      'OnTopLineChange' auslsen. }
    TopLine_neu := SendMessage(Handle, EM_GETFIRSTVISIBLELINE, 0, 0);
    if TopLine_neu<>TopLine_alt then begin
      TopLine_alt := TopLine_neu;
      TopLineChange;
    end;
    { Neue Cursorposition speichern und alten Wert in 'Pos_alt' merken. }
    if SelStart <> Pos_neu then begin
      if Pos_alt>=0 then begin
        Pos_alt := Pos_neu;
        Pos_neu := SelStart;
      end else begin
        Pos_neu := SelStart;
        Pos_alt := Pos_neu;
      end;
    end;
    inherited SelectionChange;
  end;
end;


{ ---------- Anpassung der Syntaxhervorhebung bei nderung des Inhalts }

procedure TXRichEdt.Change;
var
  Pos: Integer;
begin
  if Syntaxhervorhebung and not Coloring then begin
    Pos := SelStart;
    if Pos>Pos_alt then
      Bereich_faerben(Pos_alt, Pos-Pos_alt)
    else if Pos<Pos_alt then
      Bereich_faerben(Pos, Pos_alt-Pos);
    SelStart := Pos;
  end;
  inherited Change;
end;


procedure TXRichEdt.Loaded;
begin
  inherited Loaded;
  KeyWords1.SetRange('.txt');
  KeyWords2.SetRange('.txt');
  ClearUndo;
end;


{ -------------------- Diverse ffentliche Methoden }

{ ---------- Datei laden }

procedure TXRichEdt.LoadFromFile(const Name:string);
begin
  Visible := False;
  Lines.Clear;
  Filename := Name;
  Pos_alt := 0;
  Pos_neu := 0;
  Lines.LoadFromFile(Name);
  Bereich_faerben(0,Length(Text));
  SelStart := 0;
  ClearUndo;
  Invalidate;
  Modified := False;
  Visible := True;
end;


{ ---------- Datei speichern }

procedure TXRichEdt.SaveToFile(const Name:string);
begin
  Lines.SaveToFile(Name);
  Modified := False;
  ClearUndo;
end;


{ ---------- Wort bestimmen, auf dem der Cursor steht. }

function TXRichEdt.GetCurrentWord : string;
var
  Anfang, Ende: Integer;
  Inhalt : AnsiString;
begin
  Inhalt := Text;
  Anfang := SelStart;
  Ende := Anfang;
  while (Ende<=Length(Inhalt)) and (Inhalt[Ende] in FLetters) do Inc(Ende);
  while (Anfang>1) and (Inhalt[Anfang-1] in FLetters) do Dec(Anfang);
  if Ende-Anfang > 100 then Ende := Anfang+100;
  result := Copy(Inhalt,Anfang,Ende-Anfang);
end;


{ ---------- Cursorposition ermitteln }

function TXRichEdt.GetCaretPosition(var Col:Integer) : Integer;
begin
  result := SendMessage(Handle, EM_EXLINEFROMCHAR, 0, SelStart) + 1;
  Col := SelStart - SendMessage(Handle, EM_LINEINDEX, result-1, 0) + 1;
end;


end.

